/**
 * scannerlib.c - Code to interact with an XP100 scanner
 * through libusb.
 *
 * @authors Joe Kopena tjkopena@cs.drexel.edu 2004/06
 **/

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <math.h>

#include <usb.h>

#include "usblib.h"
#include "scannerlib.h"

#define STROBE_VENDOR   0x04a7
#define STROBE_PRODUCT  0x0427

#define DEFAULT_TIMEOUT 500

#define PAPER_IN        0x01
#define PAPER_INSERTED  0x08
#define BUTTON_DOWN     0x10

#define JUMP_TIME       200000l
#define EJECT_TIME     1400000l

#define REGISTERS_START  0x08
#define REGISTERS_END    0x7f

#define DATA_PIXEL_START    0x000b
#define SENSOR_RESOLUTION   600
#define LINE_END            0x14b4
#define FSPI                738

#define READ_LIMIT 0x40

int xp100Configure(usb_dev_handle *udev);
int xp100CheckChipset(usb_dev_handle *udev);

int xp100Jump(usb_dev_handle *udev);
int xp100ToNormal(usb_dev_handle *udev);
int xp100StartScan(usb_dev_handle *udev);
int xp100Eject(usb_dev_handle *udev);
int xp100SetTables(usb_dev_handle *udev);
int xp100SetATable(usb_dev_handle *udev, unsigned char table);
int xp100LoadSettings(char *file);
int xp100ManipulateSettings(unsigned char *buffer,
			    float widthInches,
			    float lengthInches,
			    unsigned char color,
			    unsigned char resolution,
			    unsigned char depth,
			    unsigned int *bpl,
			    unsigned int *ld);

struct usb_device *dev;

unsigned char *settings = NULL;

/*------------------------------------------------
  ------------------------------------------------*/
int sysInit(void) {

   usb_init();

   usb_find_busses();
   usb_find_devices();

   return 0;

  // end usbInit
}

void sysKill(void) {

  // end sysKill
}

/*------------------------------------------------
  ------------------------------------------------*/
usb_dev_handle *xp100Find(void) {

  struct usb_bus *bus;
  usb_dev_handle *udev = NULL;
  int vendor, product, ret;
  char string[255];

  ret = 0;
   for (bus = usb_busses; bus && !ret; bus = bus->next) {
     for (dev = bus->devices; dev; dev = dev->next) {

       vendor = dev->descriptor.idVendor;
       product = dev->descriptor.idProduct;

       if (vendor != STROBE_VENDOR || product != STROBE_PRODUCT)
	 continue;

       printf("Found the scanner on %s/%s\n", bus->dirname,
	      dev->filename);
       ret = 1;

       break;
     }
   }

   if (!dev) {
     fprintf(stderr, "Scanner was not found.\n");
     return NULL;
   }

   udev = usb_open(dev);
   if (!udev) {
     fprintf(stderr, "Could not open device.");
     return NULL;
   }

   if (dev->descriptor.iManufacturer) {
     ret = usb_get_string_simple(udev, dev->descriptor.iManufacturer,
				 string, sizeof(string));
     if (ret > 0)
       printf("- Manufacturer : %s\n", string);
     else
       printf("- Unable to fetch manufacturer string\n");
   }

   if (dev->descriptor.iProduct) {
     ret = usb_get_string_simple(udev, dev->descriptor.iProduct,
				 string, sizeof(string));
     if (ret > 0)
       printf("- Product      : %s\n", string);
     else
       printf("- Unable to fetch product string\n");
   }

   if (dev->descriptor.iSerialNumber) {
     ret = usb_get_string_simple(udev, dev->descriptor.iSerialNumber,
				 string, sizeof(string));
     if (ret > 0)
       printf("- Serial Number: %s\n", string);
     else
       printf("- Unable to fetch serial number string\n");
   }

   printf("\n");

   return udev;

  // end findXP100
}

/*------------------------------------------------*/
int xp100Init(usb_dev_handle *udev) {
  int ret;

  if ((ret = xp100Configure(udev)) < 0)
    return ret;

  // Take it back to normal (starts in reset)
  if ((ret = sendByte(udev, 0x07, 0x03)) < 0)
    return ret;

  if ((ret = xp100CheckChipset(udev)) < 0)
    return ret;

  if ((ret = xp100LoadSettings(SETTINGS_FILE)) < 0)
    return ret;

  // if ((ret = sendLog(udev, INIT_FILE)) < 0)
  // return ret;

   return 0;
  // end xp100Init
}

void xp100Kill(usb_dev_handle *udev) {

  free(settings);

  usb_close(udev);

  // end usbKill
}

/*------------------------------------------------
  ------------------------------------------------*/
int xp100Configure(usb_dev_handle *udev) {
  int ret, num;
  struct usb_interface_descriptor *interface;

   ret = usb_set_configuration(udev, 1);
   printf("  - config status: %i\n", ret);

   ret = usb_claim_interface(udev, 0);
   printf("  - claim status: %i\n", ret);

   printf("  - cycling through endpoints:\n");
   interface = &dev->config[0].interface[0].altsetting[0];
   for (num = 0; num < interface->bNumEndpoints; num++) {
     struct usb_endpoint_descriptor *endpoint;
     int address, direction, transfer_type;

     endpoint = &interface->endpoint[num]; 
     transfer_type = endpoint->bmAttributes & USB_ENDPOINT_TYPE_MASK;
     address = endpoint->bEndpointAddress & USB_ENDPOINT_ADDRESS_MASK;
     direction = endpoint->bEndpointAddress & USB_ENDPOINT_DIR_MASK;
     transfer_type = endpoint->bmAttributes & USB_ENDPOINT_TYPE_MASK;

     printf("    - endpoint %02x:", address);
     printf(" direction %s; type: ", ((direction)?"in":"out"));
     switch (transfer_type) {
     case USB_ENDPOINT_TYPE_CONTROL:
       printf("control");
       break;
     case USB_ENDPOINT_TYPE_ISOCHRONOUS:
       printf("isochronous");
       break;
     case USB_ENDPOINT_TYPE_BULK:
       printf("bulk");
       break;
     case USB_ENDPOINT_TYPE_INTERRUPT:
       printf("interrupt");
       break;
     default:
       printf("unknown");
       break;
     }
     printf("\n");

     ret = usb_clear_halt(udev, direction?address+0x80:address);
     printf("      - clear endpoint %02X status: %i\n",
	    ((direction)?address+0x80:address), ret);

     /* end cycling through endpoints */
   }

   return 0;

  // end xp100Configure
}

/*------------------------------------------------
  ------------------------------------------------*/
int xp100CheckChipset(usb_dev_handle *udev) {
  int ret;
  unsigned char val;

   ret = readRegOut(udev, 0x69, &val, 0);
   if (ret != 1) {
     printf("%s\n", usb_strerror());
     return -1;
   }

   if (val==4) {
     printf("  - chipset is LM9832 or LM9833\n\n");
   } else {
     printf("  - chipset is not correct\n\n");
     return -253;
   }

   return 0;

  // end xp100CheckChipset
}

/*------------------------------------------------
  ------------------------------------------------*/
int xp100LoadSettings(char *file) {

  unsigned int reg;
  unsigned int value;
  unsigned char *buffer;
  unsigned char *marked;
  int ret, missed, i;
  FILE *fh;

  if (!(fh = fopen(file, "rt"))) {
    fprintf(stderr, "Could not open settings file '%s'\n", file);
    return -254;
  }

  printf("  - reading %s... ", file);

  if (!(buffer = (char *)malloc(REGISTERS_END - REGISTERS_START))) {
    printf("Could not allocate settings buffer.\n");
    return -1;
  }
  if (!(marked = (char *)malloc(REGISTERS_END - REGISTERS_START))) {
    printf("Could not allocate settings buffer.\n");
    return -1;
  }

  memset(buffer, 0, REGISTERS_END - REGISTERS_START);
  memset(marked, 0, REGISTERS_END - REGISTERS_START);

  do {
    ret = 0;
    if (fscanf(fh, "%x", &reg) == EOF) {
      ret = 1;
      continue;
    }

    if (fscanf(fh, "%x", &value) == EOF) {
      printf("Register %i [%02x] did not have a matching value.\n", reg, reg);
      return -1;
      ret = -1;
      continue;
    }


    if (reg < REGISTERS_START) {
      printf("Reg %i [%02x] is out of range [%02x -- %02x]\n",
	     reg, reg, REGISTERS_START, REGISTERS_END);
      return -1;
    }

    if (reg > REGISTERS_END) {
      printf("Register %i [%02x] is out of range [%02x -- %02x]\n",
	     reg, reg, REGISTERS_START, REGISTERS_END);
      return -1;
    }

    buffer[reg - REGISTERS_START] = value;
    marked[reg - REGISTERS_START] = 1;

  } while (ret == 0);

  // scan through and looked for unmarked

  missed = 0;
  for (i = 0; i < REGISTERS_END - REGISTERS_START; i++)
    if (!marked[i]) {
      printf("    - Register %i [%02x] has not been marked.\n",
	     i + REGISTERS_START, i + REGISTERS_START);
      missed = 1;
    }

  if (missed) {
    printf("Settings file %s is missing some registers.\n", file);
    return -1;
  }

  printf("done\n\n");

  fclose(fh);

  settings = buffer;

  return 0;

  // end xp100LoadSettings
}

/*------------------------------------------------
  ------------------------------------------------*/
int xp100SetTables(usb_dev_handle *udev) {

  int ret;

  if ((ret = xp100SetATable(udev, 0x0)) < 0)
    return ret;

  if ((ret = xp100SetATable(udev, 0x1)) < 0)
    return ret;

  if ((ret = xp100SetATable(udev, 0x2)) < 0)
    return ret;

  printf("\n");

  return 0;

  // end xp100SetTables
}

int xp100SetATable(usb_dev_handle *udev, unsigned char table) {

  int ret, i, size;
  unsigned char *buffer;

  switch (table) {
  case 0: printf("  - red tables\n"); break;
  case 1: printf("  - green tables\n"); break;
  case 2: printf("  - blue tables\n"); break;
  default: printf("Unknown table %02x\n", table); return -1; break;
  }

  printf("    - setting gamma\n");
  // Blue Gamma
  if ((ret = sendByte(udev, 0x03, 0x2|(table<<2))) < 0) {
    printf("%s\n", usb_strerror());
    return ret;
  }
  if ((ret = sendByte(udev, 0x04, 0x00)) < 0) {
    printf("%s\n", usb_strerror());
    return ret;
  }
  if ((ret = sendByte(udev, 0x05, 0x00)) < 0) {
    printf("%s\n", usb_strerror());
    return ret;
  }

  size = 4096;
  if (!(buffer = malloc(size))) {
    printf("Not enough memory for blue offset table.\n");
    return -1;
  }

  // for (i = 0; i < (int)(size/2); i++)
  //  buffer[i] = 0x00;
  for (i = 0; i < size; i++)
    buffer[i] = 0xff - (unsigned char)(((int)0xff*i)/size);

  if ((ret = sendArray(udev, 0x06, buffer, size)) < 0) {
    printf("%s\n", usb_strerror());
    return ret;
  }

  free(buffer);

  printf("    - setting offset\n");
  // Blue Offset
  if ((ret = sendByte(udev, 0x03, (table<<2))) < 0) {
    printf("%s\n", usb_strerror());
    return ret;
  }
  if ((ret = sendByte(udev, 0x04, 0x00)) < 0) {
    printf("%s\n", usb_strerror());
    return ret;
  }
  if ((ret = sendByte(udev, 0x05, 0x00)) < 0) {
    printf("%s\n", usb_strerror());
    return ret;
  }

  size = 16384;
  if (!(buffer = malloc(size))) {
    printf("Not enough memory for blue offset table.\n");
    return -1;
  }

  for (i = 0; i < size; i++)
    buffer[i] = 0;

  if ((ret = sendArray(udev, 0x06, buffer, size)) < 0) {
    printf("%s\n", usb_strerror());
    return ret;
  }

  free(buffer);

  // Gain
  printf("    - setting gain\n");
  if ((ret = sendByte(udev, 0x03, 0x1|(table<<2))) < 0) {
    printf("%s\n", usb_strerror());
    return ret;
  }
  if ((ret = sendByte(udev, 0x04, 0x00)) < 0) {
    printf("%s\n", usb_strerror());
    return ret;
  }
  if ((ret = sendByte(udev, 0x05, 0x00)) < 0) {
    printf("%s\n", usb_strerror());
    return ret;
  }

  size = 16384;
  if (!(buffer = malloc(size))) {
    printf("Not enough memory for blue offset table.\n");
    return -1;
  }

  for (i = 0; i < size; i++)
    buffer[i] = 0xff;

  if ((ret = sendArray(udev, 0x06, buffer, size)) < 0) {
    printf("%s\n", usb_strerror());
    return ret;
  }

  free(buffer);

  return 0;

  // end xp100SetATable
}

/*------------------------------------------------
  ------------------------------------------------*/
int xp100WaitForPaperReady(usb_dev_handle *udev) {
  unsigned char val;
  int ret;
  unsigned char paper = 0;

     printf("\n  - waiting.\n");

     do {
       usleep(500);
       ret = readRegOut(udev, 0x02, &val, 0);
       if (ret != 1) {
	 printf("%s\n", usb_strerror());
	 return ret;
       }

       if ((val & PAPER_IN) && !paper) {
	 printf("    - paper in\n");
	 paper = 1;
       } else if (!(val & PAPER_IN) && paper) {
	 printf("    - paper out\n");
	 paper = 0;
       }

     } while (!(val & PAPER_IN) || !(val & BUTTON_DOWN));

     printf("  - moving on\n\n");

     return 0;
  // end xp100WaitForPaperReady
}

/*------------------------------------------------
  ------------------------------------------------*/
int xp100ParametricScan(usb_dev_handle *udev,
			unsigned char *outFile,
			float widthInches,
			float lengthInches,
			unsigned char color,
			unsigned char resolution,
			unsigned char depth) {

  int ret;
  unsigned int bytesPerLine, pixelsPerLine, linesOfData;
  FILE *out;
  unsigned char *buffer;
  int i;
  unsigned int lineCount = 0;
  unsigned char bpp, pk;

  switch (depth) {
  case 0: bpp = 1; break;
  case 1: bpp = 2; break;
  case 2: bpp = 4; break;
  case 3: bpp = 8; break;
  default:
    printf("Unknown depth %i\n", depth);
    return -1;
    break;
  }
  pk = (int)(pow(2, bpp)) - 1;

  if (!(buffer = malloc(REGISTERS_END - REGISTERS_START))) {
    printf("Not enough memory for local settings table.\n");
    return -1;
  }
  memcpy(buffer, settings, REGISTERS_END - REGISTERS_START);

  // Manipulate the settings
  if ((ret = xp100ManipulateSettings(buffer,
				     widthInches, lengthInches,
				     color, resolution, depth,
				     &pixelsPerLine, &linesOfData)) < 0)
    return ret;

  bytesPerLine = pixelsPerLine + 2;
  if (color) pixelsPerLine /= 3;
  pixelsPerLine *= (8 / bpp);

  // Go into reset mode here
  if ((ret = sendByte(udev, 0x18, 0x18)) < 0)
    return ret;
  if ((ret = sendByte(udev, 0x07, 0x20)) < 0)
    return ret;

  // Apply the settings
  for (i = 0; i < REGISTERS_END - REGISTERS_START; i++) {
    if ((ret = sendByte(udev, i + REGISTERS_START, buffer[i])) < 0)
      return ret;
  }

  // Come out of reset mode here
  if ((ret = sendByte(udev, 0x07, 0x00)) < 0)
    return ret;

  free(buffer);

  if((ret = xp100SetTables(udev)) < 0)
    return ret;

  /* Now we can rock
    jump
    start scan
    while paper present
      grab data
    eject
  */

  printf("  - scanning---may pause during fetch, do not stop until 'done'\n");
  printf("      [");
  fflush(stdout);

  if ((ret = xp100Jump(udev)) < 0)
    return ret;

  if ((ret = xp100StartScan(udev)) < 0)
    return ret;

  out = fopen(outFile, "wb");

  fprintf(out, "P3\n");
  fprintf(out, "%i  ", pixelsPerLine);
  fprintf(out, "%i\n", linesOfData);
  fprintf(out, "%i\n\n", pk);

  if (!(buffer = malloc(bytesPerLine))) {
    printf("Not enough memory for pixel buffer.\n");
    return -1;
  }

  do {
    int actRead, c;

    if (!(lineCount%(linesOfData/10))) {
      printf(".");
      fflush(stdout);
    }

    for (c = 0; c < bytesPerLine;) {
      if (bytesPerLine - c < READ_LIMIT)
	actRead = bytesPerLine - c;
      else actRead = READ_LIMIT;

      if ((ret = readArray(udev, 0x00, &buffer[c], actRead)) < 0) {
	printf("\n%s\n", usb_strerror());
	return -1;
      }

      c += actRead;
      // end looping through small chunks
    }

    for (i = 0; i <  bytesPerLine - 2; i++) {
      int j;

      for (j = 8-bpp; j >= 0; j-=bpp) {
	unsigned char val = pk - (unsigned char)((buffer[i] >> j) &pk);

	// fwrite(&val, 1, 1, out);
	fprintf(out, "%i ", val);
	if (!color) {
	  fprintf(out, "%i ", val);
	  fprintf(out, "%i ", val);
	  //fwrite(&val, 1, 1, out);
	  //fwrite(&val, 1, 1, out);
	}

      }

    }

    fprintf(out, "\n");

    lineCount++;

  } while (lineCount < linesOfData);

  printf("]\n");

  free(buffer);

  if((ret = xp100ToNormal(udev)) < 0)
    return ret;

  if ((ret = xp100Eject(udev)) < 0)
    return ret;

  // Turn off the lamp
  if ((ret = sendByte(udev, 0x29, 0x00)) < 0)
    return ret;

  fclose(out);

  printf("  - done.\n\n");

  return 0;

  // end xp200ParametricScan
}

/*-----------------------------------------------*/
int xp100ManipulateSettings(unsigned char *buffer,
			    float widthInches,
			    float lengthInches,
			    unsigned char color,
			    unsigned char resolution,
			    unsigned char depth,
			    unsigned int *bpl,
			    unsigned int *ld) {

  unsigned int dataPixelEnd;
  unsigned char hres;
  unsigned char colorMode = 0;
  unsigned int bytesPerLine, linesOfData;
  unsigned int b, vDPI;
  unsigned int stepSize;
  float hDPIDiv;

  // Figure out the resoluton settings
  switch (resolution) {
  case 0:
      hDPIDiv = 1;
      break;
    case 1:
      hDPIDiv = 1.5;
      break;
    case 2:
      hDPIDiv = 2;
      break;
    case 3:
      hDPIDiv = 3;
      break;
    case 4:
      hDPIDiv = 4;
      break;
  case 5:
      hDPIDiv = 6;
      break;
    case 6:
      hDPIDiv = 8;
    break;
    case 7:
      hDPIDiv = 12;
    break;
  default:
    printf("Unknown resolution setting: %i\n", resolution);
    return -1;
    break;
  }
  vDPI = SENSOR_RESOLUTION / hDPIDiv;

#define setBuffer(x,y,z) x[y - REGISTERS_START] = z;

  // Set data pixel start
  setBuffer(buffer, 0x22, (DATA_PIXEL_START>>8)&0xff);
  setBuffer(buffer, 0x23, (DATA_PIXEL_START&0xff));

  // Set data pixel end
  dataPixelEnd = (widthInches
		  * (SENSOR_RESOLUTION / 2)) + DATA_PIXEL_START;
  // dataPixelEnd = (0x06<<8) | 0xd1;
  setBuffer(buffer, 0x24, (dataPixelEnd>>8)&0xff);  // static setting: 06 d1
  setBuffer(buffer, 0x25, (dataPixelEnd&0xff));

  // Set hres
  hres = 0x40 | resolution | (depth << 3);
  setBuffer(buffer, 0x09, hres);  // static: 00

  // Figure out the color mode
  colorMode = (color) ? 0x15 : 0x14;  // color | greyscale
  setBuffer(buffer, 0x26, colorMode);  // static: 14

  // Figure out the bytes per line and number of lines
  switch (depth) {
  case 0: b = 1; break;
  case 1: b = 2; break;
  case 2: b = 4; break;
  case 3: b = 8; break;
  default:
    printf("Unknown depth %i\n", depth);
    return -1;
    break;
  }

  bytesPerLine = (2 * (int)((((dataPixelEnd - DATA_PIXEL_START)/hDPIDiv) *
			    ((color)?3:1) * b) / 16));
  linesOfData = vDPI * (lengthInches /2);

  // Set the step size
  stepSize = (LINE_END * vDPI * ((color)?3:1)) / (FSPI * 4);
  setBuffer(buffer, 0x46, (stepSize>>8)&0xff);
  setBuffer(buffer, 0x47, (stepSize&0xff));

  printf("  - scan settings:\n");
  printf("      image: %.2f, %.2f\n", widthInches, lengthInches);
  printf("      stepSize: %i [%02x %02x]\n", stepSize, (stepSize>>8)&0xff,
	 stepSize & 0xff);
  printf("      bytesPerLine: %i\n", bytesPerLine);
  printf("      linesOfData: %i\n", linesOfData);
  printf("      dataPixelStart: %i [%02x %02x]\n", DATA_PIXEL_START,
	 (DATA_PIXEL_START>>8)&0xff, DATA_PIXEL_START&0xff);
  printf("      dataPixelEnd: %i [%02x %02x]\n", dataPixelEnd,
	 (dataPixelEnd>>8)&0xff, dataPixelEnd&0xff);
  printf("      lineEnd: %i [%02x %02x]\n", LINE_END,
	 (LINE_END>>8)&0xff, LINE_END&0xff);
  printf("      colorMode: %i [%02x]\n", colorMode, colorMode);
  printf("      hres & dataMode: %i [%02x]\n", hres, hres);
  //printf("      : %i\n", );
  printf("\n");

  (*bpl) = bytesPerLine; // (72*3);
  (*ld) = linesOfData; // 2700;

  return 0;

  // end xp100ManipulateSettings
}

/*------------------------------------------------*/
int xp100Scan(usb_dev_handle *udev) {

  int ret, bytesPerLine, numLines, lineCount = 0, toRead;
  unsigned int width;
  FILE  *out;
  unsigned char *buffer;
  int i;

  /*
    jump
    start scan
    while paper present
      grab data
    eject
   */

  if ((ret = xp100Jump(udev)) < 0)
    return ret;

  if ((ret = xp100StartScan(udev)) < 0)
    return ret;

  printf("Img width=%i\n", (width = ((0x06<<8) | 0xd1) - 0x0b));
  // buffer = 

  // res = 200 dpi

  // res = 600 / 3
  // 1 bit/pixel

  out = fopen("test.pbm", "wt");

  // Setting the register
  // horz. res 09: 


  //bytesPerLine = 347+2;
  numLines =     2700; 
  bytesPerLine = (72*3) + 2;
  // numLines =     (468 * 3);

  fprintf(out, "P1\n");
  fprintf(out, "%i\n", bytesPerLine *  8);
  fprintf(out, "%i\n\n", numLines);

  toRead = bytesPerLine;
  if (!(buffer = malloc(toRead))) {
    printf("Not enough memory for pixel buffer.\n");
    return -1;
  }

  do {
    /*
    if ((ret = readRegOut(udev, 0x02, &paper, 0)) != 1) {
      printf("%s\n", usb_strerror());
      return -1;
    }

    if ((ret = readRegOut(udev, 0x01, &avail, 0)) != 1) {
      printf("%s\n", usb_strerror());
      return -1;
    }
    */

    // printf("available: %i [%02x]\n", avail, avail);

      if ((ret = readArray(udev, 0x00, buffer, toRead)) < 0) {
	printf("%s\n", usb_strerror());
	return -1;
      }
      printf("[line: %i; length: %i] ", lineCount, toRead);
      for (i = 0; i <  toRead; i++) {
	int j;
	printf("%02x", buffer[i]);
	for (j = 7; j >= 0; j--) {
	  fprintf(out, "%s", (buffer[i] & (1 << j) )?"1":"0");
	}
      }
      printf("\n");
      fprintf(out, "\n");
      lineCount++;

    // printf("  - available data: %i\n", val);
  } while (lineCount < numLines);

free(buffer);

// (paper & PAPER_IN));

  if((ret = xp100ToNormal(udev)) < 0)
    return ret;

  if ((ret = xp100Eject(udev)) < 0)
    return ret;

  if ((ret = sendByte(udev, 0x29, 0x00)) < 0) // Turn off the lamp
    return ret;

  fclose(out);

  return 0;

  // end xp100Scan
}

/*------------------------------------------------*/
int xp100Jump(usb_dev_handle *udev) {

  int ret;

  if ((ret = sendByte(udev, 0x45, 0x13)) < 0)
    return ret;

  if ((ret = sendByte(udev, 0x07, 0x01)) < 0)
    return ret;

  usleep(JUMP_TIME);

  return xp100ToNormal(udev);

  // end xp100Jump
}

/*------------------------------------------------*/
int xp100Eject(usb_dev_handle *udev) {

  int ret;
  unsigned char paper;

  if ((ret = sendByte(udev, 0x45, 0x13)) < 0)
    return ret;

  if ((ret = sendByte(udev, 0x07, 0x01)) < 0)
    return ret;

  do {
    if ((ret = readRegOut(udev, 0x02, &paper, 0)) != 1) {
      printf("%s\n", usb_strerror());
      return -1;
    }
  } while ((paper & PAPER_IN));

  usleep(EJECT_TIME);

  return xp100ToNormal(udev);

  // end xp100Jump
}

/*------------------------------------------------*/
int xp100ToNormal(usb_dev_handle *udev) {

  return sendByte(udev, 0x07, 0x00);

  // end xp100ToNormal
}

/*------------------------------------------------*/
int xp100StartScan(usb_dev_handle *udev) {

  int ret;

  if ((ret = sendByte(udev, 0x29, 0x03)) < 0)
    return ret;

  if ((ret = sendByte(udev, 0x07, 0x03)) < 0)
    return ret;

  return 0;

  // end xp100StartScan
}

/*------------------------------------------------
  ------------------------------------------------*/
int xp100ColorParametricScan(usb_dev_handle *udev,
			unsigned char *outFile,
			float widthInches,
			float lengthInches,
			unsigned char resolution,
			unsigned char depth) {

  int ret;
  unsigned int bytesPerLine, pixelsPerLine, linesOfData;
  FILE *out;
  unsigned char *buffer;
  unsigned char *buffer1;
  unsigned char *buffer2;
  unsigned char *buffer3;
  int i;
  unsigned int lineCount = 0;
  unsigned char bpp, pk;
  unsigned char color = 1;

  switch (depth) {
  case 0: bpp = 1; break;
  case 1: bpp = 2; break;
  case 2: bpp = 4; break;
  case 3: bpp = 8; break;
  default:
    printf("Unknown depth %i\n", depth);
    return -1;
    break;
  }
  pk = (int)(pow(2, bpp)) - 1;

  if (!(buffer = malloc(REGISTERS_END - REGISTERS_START))) {
    printf("Not enough memory for local settings table.\n");
    return -1;
  }
  memcpy(buffer, settings, REGISTERS_END - REGISTERS_START);

  // Manipulate the settings
  if ((ret = xp100ManipulateSettings(buffer,
				     widthInches, lengthInches,
				     color, resolution, depth,
				     &pixelsPerLine, &linesOfData)) < 0)
    return ret;

  bytesPerLine = pixelsPerLine + 2;
  if (color) pixelsPerLine /= 3;
  pixelsPerLine *= (8 / bpp);

  // Go into reset mode here
  if ((ret = sendByte(udev, 0x18, 0x18)) < 0)
    return ret;
  if ((ret = sendByte(udev, 0x07, 0x20)) < 0)
    return ret;

  // Apply the settings
  for (i = 0; i < REGISTERS_END - REGISTERS_START; i++) {
    if ((ret = sendByte(udev, i + REGISTERS_START, buffer[i])) < 0)
      return ret;
  }

  // Come out of reset mode here
  if ((ret = sendByte(udev, 0x07, 0x00)) < 0)
    return ret;

  free(buffer);

  if((ret = xp100SetTables(udev)) < 0)
    return ret;

  /* Now we can rock
    jump
    start scan
    while paper present
      grab data
    eject
  */

  printf("  - scanning---may pause during fetch, do not stop until 'done'\n");
  printf("      [");
  fflush(stdout);

  if ((ret = xp100Jump(udev)) < 0)
    return ret;

  if ((ret = xp100StartScan(udev)) < 0)
    return ret;

  out = fopen(outFile, "wb");

  fprintf(out, "P3\n");
  fprintf(out, "%i  ", pixelsPerLine);
  fprintf(out, "%i\n", linesOfData);
  fprintf(out, "%i\n\n", pk);

  if (!(buffer1 = malloc(bytesPerLine))) {
    printf("Not enough memory for pixel buffer.\n");
    return -1;
  }
  if (!(buffer2 = malloc(bytesPerLine))) {
    printf("Not enough memory for pixel buffer.\n");
    return -1;
  }
  if (!(buffer3 = malloc(bytesPerLine))) {
    printf("Not enough memory for pixel buffer.\n");
    return -1;
  }

  do {
    int actRead, c;

    if (!(lineCount%(linesOfData/10))) {
      printf(".");
      fflush(stdout);
    }

    for (c = 0; c < bytesPerLine;) {
      if (bytesPerLine - c < READ_LIMIT)
	actRead = bytesPerLine - c;
      else actRead = READ_LIMIT;

      if ((ret = readArray(udev, 0x00, &buffer1[c], actRead)) < 0) {
	printf("\n%s\n", usb_strerror());
	return -1;
      }

      c += actRead;
      // end looping through small chunks
    }

    for (c = 0; c < bytesPerLine;) {
      if (bytesPerLine - c < READ_LIMIT)
	actRead = bytesPerLine - c;
      else actRead = READ_LIMIT;

      if ((ret = readArray(udev, 0x00, &buffer2[c], actRead)) < 0) {
	printf("\n%s\n", usb_strerror());
	return -1;
      }

      c += actRead;
      // end looping through small chunks
    }

    for (c = 0; c < bytesPerLine;) {
      if (bytesPerLine - c < READ_LIMIT)
	actRead = bytesPerLine - c;
      else actRead = READ_LIMIT;

      if ((ret = readArray(udev, 0x00, &buffer3[c], actRead)) < 0) {
	printf("\n%s\n", usb_strerror());
	return -1;
      }

      c += actRead;
      // end looping through small chunks
    }

    for (i = 0; i <  bytesPerLine - 2; i++) {
      int j;

      for (j = 8-bpp; j >= 0; j-=bpp) {
	unsigned char val = pk - (unsigned char)((buffer1[i] >> j) &pk);
	fprintf(out, "%i ", val);
	val = pk - (unsigned char)((buffer3[i] >> j) &pk);
	fprintf(out, "%i ", val);
	val = pk - (unsigned char)((buffer2[i] >> j) &pk);
	fprintf(out, "%i ", val);
      }

    }

    fprintf(out, "\n");

    lineCount++;

  } while (lineCount < linesOfData);

  printf("]\n");

  free(buffer1);
  free(buffer2);
  free(buffer3);

  if((ret = xp100ToNormal(udev)) < 0)
    return ret;

  if ((ret = xp100Eject(udev)) < 0)
    return ret;

  // Turn off the lamp
  if ((ret = sendByte(udev, 0x29, 0x00)) < 0)
    return ret;

  fclose(out);

  printf("  - done.\n\n");

  return 0;

  // end xp200ColorParametricScan
}
