/**
 * usblib.h - Basic code needed to work with libusb.  Generic,
 * not XP100 specific.
 *
 * @authors Joe Kopena tjkopena@cs.drexel.edu 2004/06
 **/

#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <usb.h>

#include "usblib.h"

#define ARRAY_LENGTH  0x3c

int readRegOut(usb_dev_handle *udev,
			 unsigned char reg, unsigned char *val, int out) {

  char command[4];
  int ret;

  command[0] = 1;
  command[1] = reg;
  command[2] = 0;
  command[3] = 1;

  ret = usb_bulk_write(udev, 0x03, command, 4, 500);
  if (ret < 0)
    return ret;

  ret = usb_bulk_read(udev, 0x02, val, 1, 1000);
  if (out)
    printf("r-- %02x [01]  :  %i (%02x)  :  %i\n", reg, (int)(*val),
	   (int)(*val), ret);
  return ret;

      // readReg
}

int readReg(usb_dev_handle *udev,
		      unsigned char reg, unsigned char *val) {

  return readRegOut(udev, reg, val, 1);

      // readReg
}

int readArray(usb_dev_handle *udev,
			 unsigned char reg, unsigned char *buff, int len) {

  char command[4];
  int ret;

  command[0] = 1;
  command[1] = reg;
  command[2] = (len >> 8);
  command[3] = len & 0xff;

  ret = usb_bulk_write(udev, 0x03, command, 4, 500);
  if (ret < 0)
    return ret;

  ret = usb_bulk_read(udev, 0x02, buff, len, 1000);
  // printf("r-- %02x  :  %i\n", reg, ret);
  return ret;

      // readArray
}

int sendByte(usb_dev_handle *udev, char reg, char value) {

  char commandBuffer[5];
  commandBuffer[0] = 0;
  commandBuffer[1] = reg;
  commandBuffer[2] = 0;
  commandBuffer[3] = 1;
  commandBuffer[4] = value;

  return usb_bulk_write(udev, 0x03, commandBuffer, 5, DEFAULT_TIMEOUT);

  // end sendByte
}

int sendArray(usb_dev_handle *udev, char reg,
	      char *buffer, int len) {

  char *commandBuffer;
  int ret = 0, base, size;

  if (!(commandBuffer = (char *) malloc(ARRAY_LENGTH + 4))) {
    fprintf(stderr, "Not enough memory for command buffer.\n");
    return -255;
    // end malloc
  }

  base = 0;
  while (ret >= 0 && base < len) {
    if (base < len - ARRAY_LENGTH)
      size = ARRAY_LENGTH;
    else size = len - base;

    commandBuffer[0] = 0;
    commandBuffer[1] = reg;
    commandBuffer[2] = (size << 8);
    commandBuffer[3] = size & 0xff;

    memcpy(&commandBuffer[4], &buffer[base], size);

    // printf("    - sending %02x down pipe (%02x/%02x)\n", size+4, base, len);
    ret = usb_bulk_write(udev, 0x03, commandBuffer, size+4,
			 DEFAULT_TIMEOUT);

    base += size;
    // end looping through
  }

  free(commandBuffer);

  return ret;

  // end sendArray
}

int sendRawArray(usb_dev_handle *udev,
	      char *buffer, int len) {

  return usb_bulk_write(udev, 0x03, buffer, len, DEFAULT_TIMEOUT);

  // end sendArray
}

int sendLog(usb_dev_handle *udev, char *file) {
  unsigned int size, i, ret;
  unsigned char *buffer;
  unsigned int value;
  FILE *fh;

  if (!(fh = fopen(file, "rt"))) {
    fprintf(stderr, "Could not open log %s\n", file);
    return -254;
  }

  printf("  - writing %s...", file);

  do {
    fscanf(fh, "%x", &size);
    if (size == 0x666)
      continue;

    if (!(buffer = (char *)malloc(size))) {
      printf("Could not allocate buffer.\n");
      continue;
    }

    for (i = 0; i < size; i++) {
      fscanf(fh, "%x", &value);
      buffer[i] = value;
    }

    //printf("w-- %02x [%02x]  :  ", buffer[0], size);
    //printf("[%02x", buffer[0]);
    //for (i= 1; i < size && i < 16; i++) {
    //printf(" %02x",  buffer[i]);
    //}
    // printf("] : ");

    ret = sendRawArray(udev, buffer, size);
    // printf("%02x\n", ret);
    if (ret < 0) {
      fprintf(stderr, "%s\n", usb_strerror());
      return ret;
    }

    free(buffer);
  } while (size != 0x666);

  printf(" done\n");

  fclose(fh);

  return 0;
  // end sendLog
}
